"""
All modules that are pre-installed on the base Multyvac layer.

Generated by running the following on an empty layer:
    mods = list(pkgutil.iter_modules())
    for entry in sorted([(mod, is_pkg) for _,mod,is_pkg in mods]):
        print entry + ','
"""

# list of tuples [(package name, is a package?), ...]
modules = [
('BaseHTTPServer', False),
('Bastion', False),
('CDROM', False),
('CGIHTTPServer', False),
('Canvas', False),
('ConfigParser', False),
('Cookie', False),
('DLFCN', False),
('Dialog', False),
('DocXMLRPCServer', False),
('FileDialog', False),
('FixTk', False),
('HTMLParser', False),
('IN', False),
('MimeWriter', False),
('Queue', False),
('ScrolledText', False),
('SimpleDialog', False),
('SimpleHTTPServer', False),
('SimpleXMLRPCServer', False),
('SocketServer', False),
('StringIO', False),
('TYPES', False),
('Tix', False),
('Tkconstants', False),
('Tkdnd', False),
('Tkinter', False),
('UserDict', False),
('UserList', False),
('UserString', False),
('_LWPCookieJar', False),
('_MozillaCookieJar', False),
('__future__', False),
('_abcoll', False),
('_bsddb', False),
('_codecs_cn', False),
('_codecs_hk', False),
('_codecs_iso2022', False),
('_codecs_jp', False),
('_codecs_kr', False),
('_codecs_tw', False),
('_csv', False),
('_ctypes', False),
('_ctypes_test', False),
('_curses', False),
('_curses_panel', False),
('_elementtree', False),
('_heapq', False),
('_hotshot', False),
('_io', False),
('_json', False),
('_lsprof', False),
('_multibytecodec', False),
('_multiprocessing', False),
('_pyio', False),
('_sqlite3', False),
('_strptime', False),
('_testcapi', False),
('_threading_local', False),
('_weakrefset', False),
('abc', False),
('aifc', False),
('antigravity', False),
('anydbm', False),
('argparse', False),
('ast', False),
('asynchat', False),
('asyncore', False),
('atexit', False),
('audiodev', False),
('audioop', False),
('base64', False),
('bdb', False),
('binhex', False),
('bisect', False),
('bsddb', True),
('bz2', False),
('cProfile', False),
('calendar', False),
('cgi', False),
('cgitb', False),
('chunk', False),
('cmd', False),
('code', False),
('codecs', False),
('codeop', False),
('collections', False),
('colorsys', False),
('commands', False),
('compileall', False),
('compiler', True),
('contextlib', False),
('cookielib', False),
('copy', False),
('copy_reg', False),
('crypt', False),
('csv', False),
('ctypes', True),
('curses', True),
('datetime', False),
('dbhash', False),
('dbm', False),
('debconf', False),
('decimal', False),
('difflib', False),
('dircache', False),
('dis', False),
('distutils', True),
('doctest', False),
('dumbdbm', False),
('dummy_thread', False),
('dummy_threading', False),
('email', True),
('encodings', True),
('filecmp', False),
('fileinput', False),
('fnmatch', False),
('formatter', False),
('fpectl', False),
('fpformat', False),
('fractions', False),
('ftplib', False),
('functools', False),
('future_builtins', False),
('genericpath', False),
('getopt', False),
('getpass', False),
('gettext', False),
('glob', False),
('gzip', False),
('hashlib', False),
('heapq', False),
('hmac', False),
('hotshot', True),
('htmlentitydefs', False),
('htmllib', False),
('httplib', False),
('ihooks', False),
('imaplib', False),
('imghdr', False),
('importlib', True),
('imputil', False),
('inspect', False),
('io', False),
('json', True),
('keyword', False),
('lib2to3', True),
('linecache', False),
('linuxaudiodev', False),
('locale', False),
('logging', True),
('lsb_release', False),
('macpath', False),
('macurl2path', False),
('mailbox', False),
('mailcap', False),
('markupbase', False),
('md5', False),
('mhlib', False),
('mimetools', False),
('mimetypes', False),
('mimify', False),
('mmap', False),
('modulefinder', False),
('multifile', False),
('multiprocessing', True),
('multyvac', True),
('multyvacinit', True),
('mutex', False),
('netrc', False),
('new', False),
('nis', False),
('nntplib', False),
('ntpath', False),
('nturl2path', False),
('numbers', False),
('opcode', False),
('optparse', False),
('os', False),
('os2emxpath', False),
('ossaudiodev', False),
('parser', False),
('pdb', False),
('pickle', False),
('pickletools', False),
('pipes', False),
('pkgutil', False),
('platform', False),
('plistlib', False),
('popen2', False),
('poplib', False),
('posixfile', False),
('posixpath', False),
('pprint', False),
('profile', False),
('pstats', False),
('pty', False),
('py_compile', False),
('pyclbr', False),
('pydoc', False),
('pydoc_data', True),
('pyexpat', False),
('quopri', False),
('random', False),
('re', False),
('readline', False),
('repr', False),
('resource', False),
('rexec', False),
('rfc822', False),
('rlcompleter', False),
('robotparser', False),
('runpy', False),
('sched', False),
('sets', False),
('sgmllib', False),
('sha', False),
('shelve', False),
('shlex', False),
('shutil', False),
('site', False),
('sitecustomize', False),
('smtpd', False),
('smtplib', False),
('sndhdr', False),
('socket', False),
('sqlite3', True),
('sre', False),
('sre_compile', False),
('sre_constants', False),
('sre_parse', False),
('ssl', False),
('stat', False),
('statvfs', False),
('string', False),
('stringold', False),
('stringprep', False),
('struct', False),
('subprocess', False),
('sunau', False),
('sunaudio', False),
('symbol', False),
('symtable', False),
('sysconfig', False),
('tabnanny', False),
('tarfile', False),
('telnetlib', False),
('tempfile', False),
('termios', False),
('test', True),
('textwrap', False),
('this', False),
('threading', False),
('timeit', False),
('tkColorChooser', False),
('tkCommonDialog', False),
('tkFileDialog', False),
('tkFont', False),
('tkMessageBox', False),
('tkSimpleDialog', False),
('toaiff', False),
('token', False),
('tokenize', False),
('trace', False),
('traceback', False),
('ttk', False),
('tty', False),
('turtle', False),
('types', False),
('unittest', True),
('urllib', False),
('urllib2', False),
('urlparse', False),
('user', False),
('uu', False),
('uuid', False),
('warnings', False),
('wave', False),
('weakref', False),
('webbrowser', False),
('whichdb', False),
('wsgiref', True),
('xdrlib', False),
('xml', True),
('xmllib', False),
('xmlrpclib', False),
('zipfile', False),
]



import ast
import imp
import logging
import pkgutil

# How does this fail when we're in console mode?
# Things are imported relative to __main__
# Perhaps we'll strip out __main__ if we see it returned by cloudpickle

class ModuleDependencyAnalyzer(object):
    
    _IMP_TYPE_NAMES = {
        imp.PY_FROZEN: 'frozen',
        imp.PY_SOURCE: 'source',
        imp.PY_COMPILED: 'compiled',
        imp.C_EXTENSION: 'c-extention',
        imp.C_BUILTIN: 'built-in',
    }

    def __init__(self):
        """Creates new ModuleDependencyAnalyzer"""
        self._logger = logging.getLogger('multyvac.dependency-analyzer')
        # Root modules that have been or are being inspected
        self._inspected_modules = set()
        # Root modules that have yet to be inspected 
        self._modules_to_inspect = set()
        # Root modules that should be ignored by this (Not sent or traversed)
        self._modules_to_ignore = set()
        # Root module paths to transmit. Will never include path to modules
        # that contain c-extensions, and are thus untransmittable.
        self._paths_to_transmit = set()
        self.has_module_dependencies = False
        
    def add(self, module_name):
        """
        Adds a module to be analyzed.
        :param module_name: String of module name.
        """
        self._logger.info('Queuing module %r', module_name)
        root_module_name = self._extract_root_module(module_name)
        self._modules_to_inspect.add(root_module_name)
        while self._modules_to_inspect:
            self._inspect(self._modules_to_inspect.pop())

    def ignore(self, module_name):
        """
        Ignores modules in dependency analysis so that they are neither
        transmitted nor traversed. Should be called before dependencies
        are sent the first time.
        """
        # what if module is already part of paths to transmit?
        if hasattr(module_name, '__iter__'):
            self._modules_to_ignore.update(module_name)
        elif isinstance(module_name, str):
            self._modules_to_ignore.add(module_name)
        else:
            raise TypeError('module_name must be string')

    def get_and_clear_paths(self):
        # might be nice if this returned module names as well
        paths = self._paths_to_transmit
        if paths:
            self.has_module_dependencies = True
        self._paths_to_transmit = set()
        return paths

    def _inspect(self, root_module_name):
        """
        Determines what resources to send over (if any) for a given module.
        """
        if root_module_name in self._inspected_modules:
            self._logger.info('Already inspected module %r, skipping',
                              root_module_name)
            return
        elif root_module_name in self._modules_to_ignore:
            self._logger.info('Module %r is to be ignored, skipping',
                              root_module_name)
            return
        else:
            # Add module to set of scanned modules, before we've analyzed it
            self._inspected_modules.add(root_module_name)

        self._logger.info('Inspecting module %r', root_module_name)
        try:
            fp, pathname, description = imp.find_module(root_module_name)
        except ImportError:
            self._logger.info('Could not find module %r, skipping',
                              root_module_name)
            return
        suffix, mode, type = description
        if type == imp.PY_SOURCE:
            self._paths_to_transmit.add(pathname)
            self._logger.info('Module %r is source/compiled. Added path %r',
                              root_module_name, pathname)
            # TODO: Does this work with compiled sources?
            try:
                source_imps = self._find_imports(ast.parse(fp.read(),
                                                 root_module_name))
            except SyntaxError:
                self._logger.info('Module %r has a syntax error. '
                                  'Skipping source analysis',
                                  root_module_name)
                # For malformed source code
                source_imps = []
            # Close the file handle that's been opened for us by find_module
            fp.close()
            self._logger.info('Module %r had these imports %r',
                              root_module_name, source_imps)
            for source_imp in source_imps:
                if source_imp in self._inspected_modules:
                    self._logger.info('Module %r Source import %r '
                                      'already inspected', root_module_name,
                                      source_imp)
                elif source_imp in self._modules_to_inspect:
                    self._logger.info('Module %r Source import %r '
                                      'already queued', root_module_name,
                                      source_imp)
                elif source_imp in self._modules_to_ignore:
                    self._logger.info('Module %r Source import %r '
                                      'to be ignored', root_module_name,
                                      source_imp)
                else:
                    # Cannot be relative import since this is top-level
                    self._modules_to_inspect.add(source_imp)
                    self._logger.info('Module %r Source import %r added '
                                      'to queue', root_module_name, source_imp)
        elif type == imp.PKG_DIRECTORY:
            self._logger.info('Module %r is package. Recursing...',
                              root_module_name)
            if self._deep_inspect_path(pathname, root_module_name):
                self._paths_to_transmit.add(pathname)
                self._logger.info('Module %r has no c-extensions. Added path %r',
                                  root_module_name, pathname)
        elif type in (imp.C_EXTENSION, imp.C_BUILTIN, imp.PY_FROZEN,
                      imp.PY_COMPILED):
            self._logger.info('Module %r is %s. Skipping.',
                              root_module_name,
                             self._IMP_TYPE_NAMES[type])
        else:
            raise Exception('Unrecognized module %r type %s'
                            % (root_module_name, type))
        
    def _deep_inspect_path(self, path, package_name):
        """
        Traverses :param path: analyzing all valid Python modules.
        Returns True if this path is eligible to be sent (No c-extensions).
        Adds module references to list of modules to be inspected.
        """
        ret = True 
        for _, submodule_name, is_pkg in pkgutil.iter_modules([path]):
            self._logger.info('Inspecting submodule %r', submodule_name)
            fp, pathname, description = imp.find_module(submodule_name, [path])
            suffix, mode, type = description
            if type == imp.PY_SOURCE:
                self._logger.info('%r -> %r is source/compiled. '
                                  'Scanning imports.',
                                  package_name,
                                  submodule_name)
                # TODO: Does this work with compiled sources?
                try:
                    source_imps = self._find_imports(ast.parse(fp.read(),
                                                     submodule_name))
                except SyntaxError:
                    self._logger.info('%r -> %r has a syntax error. '
                                      'Skipping source analysis',
                                      package_name,
                                      submodule_name)
                    source_imps = []
                # Close the file handle that's been opened for us by find_module
                fp.close()
                self._logger.info('%r -> %r had these imports %r',
                                  package_name, submodule_name, source_imps)
                for source_imp in source_imps:
                    if source_imp in self._inspected_modules:
                        self._logger.info('%r -> %r -> %r already inspected',
                                          package_name,
                                          submodule_name,
                                          source_imp)
                    elif source_imp in self._modules_to_inspect:
                        self._logger.info('%r -> %r -> %r already queued',
                                          package_name,
                                          submodule_name,
                                          source_imp)
                    elif source_imp in self._modules_to_ignore:
                        self._logger.info('%r -> %r -> %r to be ignored',
                                          package_name,
                                          submodule_name,
                                          source_imp)
                    elif self._is_relative_import(source_imp, path):
                        self._logger.info('%r -> %r -> %r is relative.',
                                          package_name,
                                          submodule_name,
                                          source_imp)
                    else:
                        self._modules_to_inspect.add(source_imp)
                        self._logger.info('%r -> %r -> %r added to queue',
                                          package_name,
                                          submodule_name,
                                          source_imp)
            elif type == imp.PKG_DIRECTORY:
                self._logger.info('%r -> %r is package. Recursing...',
                                  package_name, submodule_name)
                ret = ret and self._deep_inspect_path(pathname, package_name)
            elif type in (imp.C_EXTENSION, imp.C_BUILTIN, imp.PY_FROZEN,
                          imp.PY_COMPILED):
                # TODO: Can we go from compiled Python to an AST to identify
                # imports?
                self._logger.info('%r -> %r is %s.',
                                  package_name,
                                  submodule_name,
                                  self._IMP_TYPE_NAMES[type])
                # Close the file handle that's been opened for us by find_module
                fp.close()
                ret = False
            else:
                raise Exception('Unrecognized module type %s' % submodule_name)
        
        return ret

    def _is_relative_import(self, module_name, path):
        """Checks if import is relative. Returns True if relative, False if
        absolute, and None if import could not be found."""
        try:
            # Check within the restricted path of a (sub-)package
            imp.find_module(module_name, [path])
        except ImportError:
            pass
        else:
            return True

        try:
            # Check across all of sys.path
            imp.find_module(module_name)
        except ImportError:
            pass
        else:
            return False

        # Module could not be found on system due to:
        # 1. Import that doesn't exist. "Bad import".
        # 2. Since we're only scanning the AST, there's a good chance the
        #    import's inclusion is conditional, and would never be triggered.
        #    For example, an import specific to an OS. 
        return None

    def _extract_root_module(self, module_name):
        """Given a module name, returns only the root module by ignoring 
        everything including and after the leftmost "." if one exists."""
        return module_name.split('.')[0]

    def _find_imports(self, node):
        """Recurses through AST collecting the targets of all import
        statements."""
        if isinstance(node, ast.Import):
            return {self._extract_root_module(alias.name) for alias in node.names}
        elif isinstance(node, ast.ImportFrom):
            # We ignore all imports with levels other than 0. That's because if
            # if level > 0, we know that it's a relative import, and we only
            # care about root modules.
            if node.level == 0:
                return {self._extract_root_module(node.module)}
            else:
                return set()
        elif hasattr(node, 'body') and hasattr(node.body, '__iter__'):
            # Not all bodies are lists (for ex. exec)
            imps = set()
            for child_node in node.body:
                imps.update(self._find_imports(child_node))
            return imps
        else:
            return set()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    mda = ModuleDependencyAnalyzer()
    import preinstalls
    mda.ignore([module for module, pkg in preinstalls.modules])
    #mda.ignore('encodings')
    #mda.ignore('test')
    mda.add('tau')
    #mda.ignore('s')
    #mda.add('s')
    #mda.add('numpy')
    print 'inspected modules', mda._inspected_modules
    print 'modules to inspect', mda._modules_to_inspect
    print 'paths to trans', mda._paths_to_transmit



from .multyvac import Multyvac

# Default Multyvac singleton
_multyvac = Multyvac()

on_multyvac = _multyvac.on_multyvac
send_log_to_support = _multyvac.send_log_to_support

# Job methods that have been elevated to top level
modulemgr = _multyvac.job._modulemgr
get = _multyvac.job.get
get_by_name = _multyvac.job.get_by_name
list = _multyvac.job.list
kill = _multyvac.job.kill
shell_submit = _multyvac.job.shell_submit
submit = _multyvac.job.submit

# All other modules
config = _multyvac.config
volume = _multyvac.volume
api_key = _multyvac.api_key
layer = _multyvac.layer



import datetime
import logging
import os
import random
import time

try:
    from cloghandler import (
        ConcurrentRotatingFileHandler as RotatingFileHandler
    )
except ImportError:
    from logging.handlers import RotatingFileHandler 

import requests

class MultyvacError(Exception):
    pass

class MultyvacRequestError(Exception):
    """Exception class for errors when making web requests to the
    Multyvac API."""

    def __init__(self, code, message, hint=None, retry=False):
        Exception.__init__(self, code, message, hint, retry)
        self.code = code
        self.message = message
        self.hint = hint
        self.retry = retry
    
    def __str__(self):
        return '%s (Code: %s Hint: %s)' % (self.message, self.code, self.hint)
    
    def __repr__(self):
        return 'MultyvacError({code}, "{message}", {hint})'.format(
                    code=self.code,
                    message=self.message,
                    hint=self.hint,
                )

class Multyvac(object):
    """
    Multyvac

    The primary object for interacting with the Multyvac API.
    All Multyvac modules are exposed through this.
    """
    
    _ASK_GET = 'GET'
    _ASK_POST = 'POST'
    _ASK_PUT = 'PUT'
    
    def __init__(self, api_key=None, api_secret_key=None, api_url=None):
        self._session = requests.session()
        
        from .config import ConfigModule
        # Note: At this time, the rest of the Multyvac modules have not been
        # initialized. So the constructor should not do anything that requires
        # any other modules (ie. Do not use the ApiKey module).
        self.config = ConfigModule(self, api_key, api_secret_key, api_url)
        
        # Must be after config
        self._setup_logger()
        
        from .job import JobModule
        self.job = JobModule(self)
        from .layer import LayerModule
        self.layer = LayerModule(self)
        from .volume import VolumeModule
        self.volume = VolumeModule(self)
        from .api_key import ApiKeyModule
        self.api_key = ApiKeyModule(self)
    
    def _setup_logger(self):
        """
        Sets up a rotating file logger.
        TODO: Have config option for printing to screen.
        """
        
        logs_path = os.path.join(self.config.get_multyvac_path(), 'log')
        if not os.path.exists(logs_path):
            self.config._create_path_ignore_existing(logs_path)
        log_path = os.path.join(logs_path, 'multyvac.log')
            
        self._logger = logging.getLogger('multyvac')
        self._logger.setLevel(logging.INFO)
    
        handler = RotatingFileHandler(log_path, 'a', 512*1024, 10)
        formatter = logging.Formatter(
            '[%(asctime)s] - [%(levelname)s] - %(name)s: %(message)s'
        )
        handler.setFormatter(formatter)
        self._logger.addHandler(handler)
    
    def _get_session_method(self, method):
        if method == self._ASK_POST:
            return self._session.post
        elif method == self._ASK_GET:
            return self._session.get
        elif method == self._ASK_PUT:
            return self._session.put
        else:
            raise KeyError('Unknown method "%s"' % method)
    
    def _ask(self, method, uri, auth=None, params=None, data=None,
             headers=None, files=None):
        
        attempt = 0
        max_attempts = 4
        while True:
            self._logger.info('%s request to %s with params %r',
                              method, uri, params)
            try:
                r = self._ask_helper(method,
                                     uri,
                                     auth=auth,
                                     params=params,
                                     data=data,
                                     headers=headers,
                                     files=files)
                return r
            except MultyvacRequestError as e:
                attempt += 1
                if e.retry and attempt < max_attempts:
                    delay = 2**attempt * random.random()
                    self._logger.info('Request failed. Retrying in %.1fs',
                                      delay)
                    time.sleep(delay)
                    continue
                else:
                    raise
    
    def _ask_helper(self, method, uri, auth, params, data, headers, files):
        
        if not auth:
            auth = self.config.get_auth()
        
        r = self._get_session_method(method)(
                self.config.api_url + uri,
                auth = auth,
                params=params,
                data=data,
                headers=headers,
                files=files,
            )
        
        try:
            obj = r.json()
        except ValueError:
            raise MultyvacRequestError(500,
                                       'Unknown error',
                                       hint=r.text)
        if 'error' in obj:
            raise MultyvacRequestError(obj['error']['code'],
                                       obj['error']['message'],
                                       obj['error'].get('hint'),
                                       obj['error'].get('retry'))
        
        return obj
        
    def on_multyvac(self):
        """Returns True if this process is currently running on Multyvac."""
        return os.environ.get('ON_MULTYVAC') == 'true'
    
    def send_log_to_support(self):
        """Sends this machine's log file to Multyvac support."""
        
        log_path = os.path.join(self.config.get_multyvac_path(),
                                'log/multyvac.log')
        if os.path.exists(log_path):
            with open(log_path) as f:
                files = {'file': ('multyvac.log', f.read())}
            self._ask(self._ASK_POST,
                      '/report/client_log/',
                      files=files)
            return True
        else:
            return False

class MultyvacModule(object):
    """All modules should extend this class."""
    def __init__(self, multyvac):
        self.multyvac = multyvac
        logger_name = self.__class__.__name__.lower()[:-len('module')]
        self._logger = logging.getLogger('multyvac.%s' % logger_name)

    @staticmethod
    def clear_null_entries(d):
        for k, v in d.items():
            if v is None:
                del d[k]

    @staticmethod
    def convert_str_to_datetime(s):
        try:
            return datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S.%f')
        except ValueError:
            # FIXME
            return datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S')

    @staticmethod
    def check_success(r):
        return r['status'] == 'ok'
    
    @staticmethod
    def is_iterable_list(obj):
        return hasattr(obj, '__iter__')


class MultyvacModel(object):
    def __init__(self, multyvac=None, **kwargs):
        if multyvac:
            self.multyvac = multyvac
        else:
            raise Exception('Needs multyvac object for now')
            
    def __str__(self):
        return repr(self)




import base64
try: import cPickle as pickle
except: import pickle
try: from cStringIO import StringIO
except: import StringIO
import json
import socket
import subprocess
import time

from .multyvac import (
    Multyvac,
    MultyvacModel,
    MultyvacModule,
    MultyvacRequestError,
)
from .util import preinstalls
from .util.module_dependency import ModuleDependencyAnalyzer

class Job(MultyvacModel):
    
    status_waiting = 'waiting'
    status_queued = 'queued'
    status_processing = 'processing'
    
    status_done = 'done'
    status_error = 'error'
    status_killed = 'killed'
    status_stalled = 'stalled'
    
    finished_statuses = [status_done,
                         status_error,
                         status_killed,
                         status_stalled,
                         ]
    
    def __init__(self, jid, **kwargs):
        MultyvacModel.__init__(self, **kwargs)
        self.jid = jid
        
        self.cmd = kwargs.get('cmd')
        self.core = kwargs.get('core')
        self.created = kwargs.get('created_at')
        self.multicore = kwargs.get('multicore')
        self.name = kwargs.get('tags', {}).get('name')
        self.status = kwargs.get('status')
        self.tags = kwargs.get('tags')
    
        result = kwargs.get('result')
        self.result_type = kwargs.get('result_type')
        if result and self.result_type  == 'pickle':
            self.result = pickle.loads(base64.b64decode(result))
        elif result and self.result_type == 'binary':
            self.result = base64.b64decode(result)
        else:
            self.result = result
        self.return_code = kwargs.get('return_code')
        
        self.started_at = kwargs.get('started_at')
        self.finished_at = kwargs.get('finished_at')
        self.runtime = kwargs.get('runtime')
        
        self.cputime_user = kwargs.get('collected', {}).get('cputime_user')
        self.cputime_system = kwargs.get('collected', {}).get('cputime_system')
        self.memory_failcnt = kwargs.get('collected', {}).get('memory_failcnt')
        self.memory_max_usage = kwargs.get('collected', {}).get('memory_max_usage')
        self.ports = kwargs.get('collected', {}).get('ports')
        self.stderr = kwargs.get('stderr')
        self.stdout = kwargs.get('stdout')
    
    def kill(self):
        """
        :param jid: Can be a number or list of numbers.
        """
        
        data = {'jid': self.jid}
        r = self.multyvac._ask(Multyvac._ASK_POST,
                               '/job/kill',
                               data=data)
        return MultyvacModule.check_success(r)
    
    def wait(self, status=finished_statuses, timeout=None):
        """
        Wait for the job to reach the specified status.
        
        :param status: Can be a status string or a list of strings.
        
        Returns the status if the job reaches it, otherwise on a timeout
        returns False.
        """
        start_time = time.time()
        poll_period = 1
        max_poll_period = 10
        while True:
            current_time = time.time()
            time_elapsed = current_time - start_time
            if timeout and time_elapsed > timeout:
                return False
            elif self.status == status or self.status in status:
                return self.status
            else:
                sleep_time = (min(poll_period, timeout - time_elapsed)
                              if timeout else poll_period)
                time.sleep(sleep_time)
                if poll_period < max_poll_period:
                    poll_period += 0.5
                self.update()
    
    def wait_for_open_port(self, port, timeout=None):
        """Blocks until it detects that port has been opened by the job."""
        self.wait(self.finished_statuses + [self.status_processing], timeout)
        if self.status == self.status_processing:
            attempt = 1
            max_attempts = 5
            while attempt < max_attempts and self.status == Job.status_processing:
                if not self.ports or not self.ports.get('tcp', {}).get(str(port)):
                    # Wait for SSH to start
                    time.sleep(2.0)
                    attempt += 1
                    self.update()
                else:
                    return True
            else:
                return False
        else:
            return False
    
    def open_ssh_console(self):
        """Opens an SSH console to a running job. If a job is queued, waits
        indefinitely until it is processing."""
        
        if self.wait_for_open_port(22):
            info = self.ports.get('tcp', {}).get('22')
            address = info['address']
            port = info['port']
            cmd = ('ssh -o UserKnownHostsFile=/dev/null '
                   '-o StrictHostKeyChecking=no -X -p {port} -i {key_path} '
                   ' multyvac@{address}'.format(
                        port=port,
                        key_path=self.multyvac.config.path_to_private_key(),
                        address=address,)
                   )
            p = subprocess.Popen(cmd, shell=True)
            p.wait()
        
    def run_command(self, cmd):
        """
        Runs the specified command over ssh.
        
        If successful, returns (stdout, stderr). Otherwise, returns None.
        """
        self.wait(self.finished_statuses + [self.status_processing])
        if self.status == self.status_processing:
            if not self.ports.get('tcp', {}).get('22'):
                # Wait for SSH to start
                time.sleep(8.0)
                self.update()
            info = self.ports.get('tcp', {}).get('22')
            if not info:
                self.multyvac.job._logger.info('Port 22 is not open')
                return
            address = info['address']
            port = info['port']
            cmd = ('ssh -o UserKnownHostsFile=/dev/null '
                   '-o StrictHostKeyChecking=no -p {port} -i {key_path} '
                   ' multyvac@{address} {cmd}'.format(
                        port=port,
                        key_path=self.multyvac.config.path_to_private_key(),
                        address=address,
                        cmd=cmd)
                   )
            p = subprocess.Popen(cmd,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 shell=True)
            return p.communicate()
        else:
            self.multyvac.job._logger.info('Cannot SSH into finished job')
            return
    
    def update(self):
        """Updates this Job object with the latest version available of itself
        from Multyvac."""
        j = self.multyvac.job.get(self.jid)
        self.__dict__ = j.__dict__
        return self.status

    def __repr__(self):
        if self.name:
            return 'Job(%s, name=%s)' % (repr(self.jid), repr(self.name))
        else:
            return 'Job(%s)' % repr(self.jid)


class JobModule(MultyvacModule):

    def __init__(self, *args, **kwargs):
        MultyvacModule.__init__(self, *args, **kwargs)
        self._modulemgr = ModuleDependencyAnalyzer()
        preinstalled_modules = [name for name, _ in preinstalls.modules]
        self._modulemgr.ignore(preinstalled_modules)
    
    def _normalize_vol(self, vol):
        if isinstance(vol, basestring):
            return [{'name': vol}]
        elif isinstance(vol, dict):
            return [vol]
        elif hasattr(vol, '__iter__'):
            if isinstance(vol[0], dict):
                return vol
            elif isinstance(vol[0], basestring):
                return [{'name': v} for v in vol]
            else:
                raise TypeError('vol must be a iterable of strings or dicts')
        else:
            raise TypeError('vol must be a string, dict, or iterable')
        
    def _normalize_layer(self, layer):
        if isinstance(layer, basestring):
            return [{'name':layer}]
        elif isinstance(layer, dict):
            return [layer]
        else:
            raise TypeError('layer must be a string or dict')

    def shell_submit(self, cmd, _name=None, _core='c1', _multicore=1,
                     _layer=None,  _vol=None, _env=None,
                     _result_source='stdout', _result_type='binary',
                     _max_runtime=None, _profile=False, _restartable=True,
                     _tags=None, _depends_on=None, _stdin=None):
        
        job = {
               'cmd': cmd,
               'name': _name,
               'core': _core,
               'multicore': _multicore,
               'profile': _profile,
               'restartable': _restartable,
               'depends_on': _depends_on,
               'tags': _tags,
               'layer': self._normalize_layer(_layer) if _layer else None,
               'vol': self._normalize_vol(_vol) if _vol else None,
               'env': _env,
               'result_source': _result_source,
               'result_type': _result_type,
               'max_runtime': _max_runtime,
               }
        
        if _stdin:
            job['stdin'] = base64.b64encode(_stdin)
        
        MultyvacModule.clear_null_entries(job)
        
        payload = {'jobs': [job]}
        
        headers = {'content-type': 'application/json'}
        
        r = self.multyvac._ask(Multyvac._ASK_POST,
                               '/job',
                               data=json.dumps(payload),
                               headers=headers)
        return r['jids'][0]

    def _get_auto_module_volume_name(self):
        return 'auto-deps-%s' % socket.gethostname()

    def submit(self, f, *args, **kwargs):
        """Submit job based on Python function to Multyvac.
        See arguments to shell_submit."""
        #_name=None, _core='c1', _multicore=1, _layer=None, _vol=None,
        #       _env=None, _result_source='stdout', _max_runtime=None,
        #       _profile=False, _restartable=True, _tags=None, _depends_on=None, _stdin=None):
        
        f_kwargs = {}
        for k, v in kwargs.items():
            if not k.startswith('_'):
                f_kwargs[k] = v
                del kwargs[k]
        
        from .util.cloudpickle import CloudPickler
        
        s = StringIO()
        cp = CloudPickler(s, 2)
        cp.dump((f, args, f_kwargs))
        
        # Add modules
        for module in cp.modules:
            self._modulemgr.add(module.__name__)
        
        mod_paths = self._modulemgr.get_and_clear_paths()
        
        vol_name = self._get_auto_module_volume_name()
        if self._modulemgr.has_module_dependencies:
            v = self.multyvac.volume.get(vol_name)
            if not v:
                try:
                    self.multyvac.volume.create(vol_name, '/pymodules')
                except MultyvacRequestError as e:
                    if 'name already exists' not in e.message:
                        raise
                v = self.multyvac.volume.get(vol_name)
            if mod_paths:
                v.sync_up(mod_paths, '')
            
        kwargs['_stdin'] = s.getvalue()
        kwargs['_result_source'] = 'file:/tmp/.result'
        kwargs['_result_type'] = 'pickle'
        if self._modulemgr.has_module_dependencies:
            kwargs.setdefault('_vol', []).append(vol_name)
        # Add to the PYTHONPATH if user is using it as well
        env = kwargs.setdefault('_env', {})
        if env.get('PYTHONPATH'):
            env['PYTHONPATH'] = env['PYTHONPATH'] + ':/pymodules'
        else:
            env['PYTHONPATH'] = '/pymodules'
        
        return self.shell_submit(
            'python -m multyvacinit.pybootstrap',
            **kwargs
        )

    def list(self, jid=None, name=None, limit=50):
        """Returns a list of jobs matching.
        Add filter by tags."""
        params = {'jid': jid,
                  'name': name,
                  'limit': limit,
                  #'field': self.fields,
                  }
        MultyvacModule.clear_null_entries(params)
        
        r = self.multyvac._ask(Multyvac._ASK_GET,
                               '/job',
                               params=params)
        for job in r['jobs']:
            job['created_at'] = MultyvacModule.convert_str_to_datetime(job['created_at'])
        return r['jobs']
    
    def get(self, jid):
        """
        :param jid: Can be a number or list of numbers.
        
        Returns a list if input was a list, otherwise returns an object.
        """
        iter_in = MultyvacModule.is_iterable_list(jid)
        params = {'jid': jid}
        if not iter_in:
            params['limit'] = 1
        jobs = self._get(params)
        if iter_in:
            return jobs
        else:
            return jobs[0] if jobs else None
    
    def get_by_name(self, name):
        """Returns the job with the matching :param name:.
        
        This function cannot be used to query multiple names are once."""
        
        iter_in = MultyvacModule.is_iterable_list(name)
        if iter_in:
            raise ValueError('name can only be a string, not a list')
        
        params = {'name': name,
                  'limit': 1}
        jobs = self._get(params)
        
        return jobs[0] if jobs else None
    
    def _get(self, more_params):
        
        params = {#'field': self.fields
                  }
        params.update(more_params)
        
        r = self.multyvac._ask(Multyvac._ASK_GET,
                               '/job',
                               params=params)
        for job in r['jobs']:
            job['created_at'] = MultyvacModule.convert_str_to_datetime(job['created_at'])
        return [Job(multyvac=self.multyvac, **job) for job in r['jobs']]

    def kill(self, jid):
        """
        :param jid: Can be a number or list of numbers.
        """
        
        data = {'jid': jid}
        r = self.multyvac._ask(Multyvac._ASK_POST,
                               '/job/kill',
                               data=data)
        return MultyvacModule.check_success(r)



import base64
import json

from .multyvac import (
    Multyvac,
    MultyvacModel,
    MultyvacModule,
)

class Layer(MultyvacModel):
    """Represents a Multyvac Layer and its associated operations."""
    
    def __init__(self, name, **kwargs):
        """Creates a new layer."""
        MultyvacModel.__init__(self, **kwargs)
        
        self.name = name
        self.size = kwargs.get('size')
        self.created_at = kwargs.get('created_at')
    
    def mkdir(self, path):
        """Creates a new directory at :param path:."""
        r = self.multyvac._ask(Multyvac._ASK_PUT,
                               '/layer/%s/mkdir' % self.name,
                               params={'path': path},
                               )
        return MultyvacModule.check_success(r)

    def put_contents(self, contents, target_path, target_mode=None):
        """
        Creates a new file with the specified contents.
        :param contents: A string of the contents of the new file.
        :param target_path: The path in the volume to create the new file.
        :param target_mode: The mode in octal notation of the new file.
            Ex. 0755.
        """
        files = {'file': (target_path, contents)}
        data = {'file_mode': target_mode}
        r = self.multyvac._ask(Multyvac._ASK_PUT,
                               '/layer/%s' % self.name,
                               files=files,
                               data=data,
                               )
        return MultyvacModule.check_success(r)

    def get_contents(self, path):
        """
        Returns the contents of the file as a string.
        :param path: The path to the file in this layer.
        """
        r = self.multyvac._ask(Multyvac._ASK_GET,
                               '/layer/%s' % self.name,
                               params={'path': [path]},
                               )
        f = r['files'][0]
        f['contents'] = base64.b64decode(f['contents'])
        return f
    
    def get_file(self, remote_path, local_path):
        """
        Copies the file at :param remote_path: in the layer to the
        :param local_path:.
        """
        # FIXME: Make more memory efficient
        f = self.get_contents(remote_path)
        with open(local_path, 'w') as target_f:
            target_f.write(f['contents'])
    
    def put_file(self, local_path, remote_path, target_mode=None):
        """
        Copies the file at :param local_path: to the layer's
        :param remote_path:.
        """
        # FIXME: Make more memory efficient
        with open(local_path) as f:
            self.put_contents(f.read(), remote_path, target_mode)
    
    def ls(self, path):
        """
        Lists the contents of :param path:.
        
        Returns a list of dicts. Each dict specifies the path to an element,
        the mode, the size, and the type of element, file (f) or directory (d).
        """
        r = self.multyvac._ask(Multyvac._ASK_GET,
                               '/layer/%s/ls' % self.name,
                               params={'path': path},
                               )
        return r['ls']
    
    def rm(self, path):
        """Remove file or directory from the layer"""
        # TODO: Support recursive flag
        r = self.multyvac._ask(Multyvac._ASK_POST,
                               '/layer/%s/rm' % self.name,
                               params={'path': path},
                               )
        return MultyvacModule.check_success(r)

    def create_modify_job(self, vol=None, max_runtime=3600):
        """
        Creates a job that can be SSH-ed into. You can SSH into this job,
        and any modifications you make to the filesystem will become a
        permanent part of the layer.
        
        Returns a job_id. You'll want to run layer.ssh(job_id) to open an SSH
        console to do things like "apt-get". You can also use sudo for root
        access.
        
        Be sure to call layer.done_modifying(job_id) after you're done to kill
        the job.
        
        :param vol: The names of any volumes you want to have mounted in the
            new job.
        :param max_runtime: The number of seconds before the modification job
            is forcibly killed. This helps prevent runaway jobs.
        """
        jid = self.multyvac.job.shell_submit(
            'sleep %s' % max_runtime,
            _name='layer modify (%r)' % self.name,
            _vol=vol,
            _layer={'name': self.name, 'mount_rw': True},
        )
        return jid
    
    def ssh_into_modify_job(self, jid):
        job = self.multyvac.job.get(jid)
        if not job.wait_for_open_port(22):
            raise Exception('Job did not open port 22')
        job.open_ssh_console()
    
    def done_modifying(self, jid):
        job = self.multyvac.job.get(jid)
        job.run_command('killall -10 python')

    def __repr__(self):
        return "Layer(%s)" % repr(self.name)

class LayerModule(MultyvacModule):
    
    def get(self, name):
        """Returns the volume with :param name:."""
        ls = self.list(name)
        if ls:
            return ls[0]
    
    def create(self, name):
        """
        Creates a new layer.
        
        :param name: The name of the volume. Must not already exist.
        """
        layer = {'name': name,
                 }
        MultyvacModule.clear_null_entries(layer)
        payload = {'layer': layer}
        headers = {'content-type': 'application/json'}
        r = self.multyvac._ask(Multyvac._ASK_POST,
                               '/layer',
                               data=json.dumps(payload),
                               headers=headers)
        return MultyvacModule.check_success(r)

    def list(self, name=None):
        """
        Returns a list of layer objects.
        
        :param name: A string or list of strings to filter results to only a
            set of layers. 
        """
        params = {}
        if name:
            params['name'] = name
        r = self.multyvac._ask(Multyvac._ASK_GET,
                               '/layer',
                               params=params)
        for layer in r['layers']:
            layer['created_at'] = MultyvacModule.convert_str_to_datetime(layer['created_at'])
        return [Layer(multyvac=self.multyvac, **v) for v in r['layers']]
